import { __awaiter, __decorate, __generator, __metadata, __values } from "tslib";
import { CanvasConfig, DisplayObjectPool, RenderingPluginContribution, OffscreenCanvasCreator, Point } from '@antv/g';
import { PathGeneratorFactory, RBushRoot } from '@antv/g-plugin-canvas-renderer';
import { mat4, vec3 } from 'gl-matrix';
import { inject, singleton, Syringe } from 'mana-syringe';
export var PointInPathPickerFactory = Syringe.defineToken('PointInPathPicker');
/**
 * pick shape(s) with Mouse/Touch event
 *
 * 1. find AABB with r-tree
 * 2. do math calculation with geometry in an accurate way
 */

var CanvasPickerPlugin =
/** @class */
function () {
  function CanvasPickerPlugin() {
    var _this = this;

    this.isHit = function (displayObject, position, worldTransform) {
      // use picker for current shape's type
      var pick = _this.pointInPathPickerFactory(displayObject.nodeName);

      if (pick) {
        // invert with world matrix
        var invertWorldMat = mat4.invert(mat4.create(), worldTransform); // transform client position to local space, do picking in local space

        var localPosition = vec3.transformMat4(vec3.create(), vec3.fromValues(position[0], position[1], 0), invertWorldMat); // account for anchor

        var halfExtents = displayObject.getGeometryBounds().halfExtents;
        var anchor = displayObject.parsedStyle.anchor;
        localPosition[0] += (anchor && anchor[0].value || 0) * halfExtents[0] * 2;
        localPosition[1] += (anchor && anchor[1].value || 0) * halfExtents[1] * 2;

        if (pick(displayObject, new Point(localPosition[0], localPosition[1]), _this.isPointInPath)) {
          return true;
        }
      }

      return false;
    };
    /**
     * use native picking method
     * @see https://developer.mozilla.org/zh-CN/docs/Web/API/CanvasRenderingContext2D/isPointInPath
     */


    this.isPointInPath = function (displayObject, position) {
      var context = _this.offscreenCanvas.getOrCreateContext(_this.canvasConfig.offscreenCanvas);

      var generatePath = _this.pathGeneratorFactory(displayObject.nodeName);

      if (generatePath) {
        generatePath(context, displayObject.parsedStyle);
      }

      return context.isPointInPath(position.x, position.y);
    };
  }

  CanvasPickerPlugin.prototype.apply = function (renderingService) {
    var _this = this;

    renderingService.hooks.pick.tapPromise(function (result) {
      return __awaiter(_this, void 0, void 0, function () {
        var topmost, _a, x, y, position, rBushNodes, queriedIds, hitTestList, pickedDisplayObjects, hitTestList_1, hitTestList_1_1, displayObject, clipPath, objectToTest, worldTransform;

        var e_1, _b;

        var _this = this;

        return __generator(this, function (_c) {
          topmost = result.topmost, _a = result.position, x = _a.x, y = _a.y;
          position = vec3.fromValues(x, y, 0);
          rBushNodes = this.rBush.search({
            minX: position[0],
            minY: position[1],
            maxX: position[0],
            maxY: position[1]
          });
          queriedIds = rBushNodes.map(function (node) {
            return node.id;
          });
          hitTestList = [];
          rBushNodes.forEach(function (_a) {
            var id = _a.id;

            var displayObject = _this.displayObjectPool.getByEntity(id);

            if (displayObject.isVisible() && !displayObject.isCulled() && displayObject.isInteractive()) {
              // parent is not included, eg. parent is clipped
              if (displayObject.parentNode && queriedIds.indexOf(displayObject.parentNode.entity) === -1) {
                return;
              }

              hitTestList.push(displayObject);
            }
          }); // find group with max z-index

          hitTestList.sort(function (a, b) {
            return b.sortable.renderOrder - a.sortable.renderOrder;
          });
          pickedDisplayObjects = [];

          try {
            for (hitTestList_1 = __values(hitTestList), hitTestList_1_1 = hitTestList_1.next(); !hitTestList_1_1.done; hitTestList_1_1 = hitTestList_1.next()) {
              displayObject = hitTestList_1_1.value;
              clipPath = displayObject.parsedStyle.clipPath;
              objectToTest = clipPath || displayObject;
              worldTransform = displayObject.getWorldTransform(); // clipped's world matrix * clipPath's local matrix

              if (clipPath) {
                worldTransform = mat4.multiply(mat4.create(), worldTransform, clipPath.getLocalTransform());
              }

              if (this.isHit(objectToTest, position, worldTransform)) {
                if (topmost) {
                  result.picked = [displayObject];
                  return [2
                  /*return*/
                  , result];
                } else {
                  pickedDisplayObjects.push(displayObject);
                }
              }
            }
          } catch (e_1_1) {
            e_1 = {
              error: e_1_1
            };
          } finally {
            try {
              if (hitTestList_1_1 && !hitTestList_1_1.done && (_b = hitTestList_1.return)) _b.call(hitTestList_1);
            } finally {
              if (e_1) throw e_1.error;
            }
          }

          result.picked = pickedDisplayObjects;
          return [2
          /*return*/
          , result];
        });
      });
    });
  };

  __decorate([inject(DisplayObjectPool), __metadata("design:type", DisplayObjectPool)], CanvasPickerPlugin.prototype, "displayObjectPool", void 0);

  __decorate([inject(CanvasConfig), __metadata("design:type", Object)], CanvasPickerPlugin.prototype, "canvasConfig", void 0);

  __decorate([inject(OffscreenCanvasCreator), __metadata("design:type", OffscreenCanvasCreator)], CanvasPickerPlugin.prototype, "offscreenCanvas", void 0);

  __decorate([inject(RBushRoot), __metadata("design:type", Function)], CanvasPickerPlugin.prototype, "rBush", void 0);

  __decorate([inject(PathGeneratorFactory), __metadata("design:type", Function)], CanvasPickerPlugin.prototype, "pathGeneratorFactory", void 0);

  __decorate([inject(PointInPathPickerFactory), __metadata("design:type", Function)], CanvasPickerPlugin.prototype, "pointInPathPickerFactory", void 0);

  CanvasPickerPlugin = __decorate([singleton({
    contrib: RenderingPluginContribution
  })], CanvasPickerPlugin);
  return CanvasPickerPlugin;
}();

export { CanvasPickerPlugin };